/*
 * Author: Petter Reinholdtsen <pere@td.org.uit.no>
 * Date:   2000-03-19
 *
 * EyeCam (VV6300) Video for Linux driver.
 *
 * More info and latest version is available from
 * <URL:http://www.ee.uwa.edu.au/~braunl/eyebot/>.
 *
 * The Linux parport API is documented on
 * <URL:http://www.cyberelk.demon.co.uk/parport/parport-doc>
 * <URL:http://www.cyberelk.demon.co.uk/parport/parport-lowlevel.txt>
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/parport.h>
#include <linux/sched.h>
#include <linux/version.h>
#include <linux/videodev.h>
#include <asm/uaccess.h>

#define VID_HARDWARE_EYECAM (10000)
#define EYECAM "EyeCam"

struct eyecam_device
{
  struct video_device vdev;
  struct pardevice *pdev;
  struct parport *pport;
  int users;
  struct wait_queue *capture_wait;
  int capture_ready;
  struct video_buffer capture_fb;

  unsigned int framecount;
  unsigned int framesize;
  u8 *framebuf;

  unsigned int stat_irq;
  unsigned int stat_autofd;
  unsigned int stat_status;
  unsigned int stat_control;
};

#define MAX_CAMS 4
static struct eyecam_device *eyecams[MAX_CAMS];
static unsigned int eyecam_num_cams = 0;

static int hardware_restart_capture() { return 0; };
static int hardware_brightness() { return 0; };
static int hardware_saturation() { return 0; };
static int hardware_hue() { return 0; };
static int hardware_contrast() { return 0; };
static int hardware_set_fb() { return 0; };
static int hardware_set_hue() { return 0; };
static int hardware_set_saturation() { return 0; };
static int hardware_set_contrast() { return 0; };
static int hardware_set_brightness() { return 0; };
static int hardware_detect(struct eyecam_device *eyecam);
static int hardware_setup() { return 0; };
static int hardware_capture_on() { return 0; };
static int hardware_capture_off() { return 0; };

int int2str(char *dst, int value, int base, int len, int zerofill);

static void
eyecam_init_stat(struct eyecam_device *eyecam)
{
  eyecam->stat_irq = 0;
  eyecam->stat_autofd = 0;
}
static void
eyecam_print_stat(struct eyecam_device *eyecam)
{
  char sbuf[9], cbuf[9];
  printk(KERN_ERR "eyecam: Statistics dump:\n");
  int2str(sbuf, parport_read_status(eyecam->pport), 2, 8, 1);
  int2str(cbuf, parport_read_control(eyecam->pport), 2, 8, 1);
  printk(KERN_ERR "eyecam: s=%s c=%s #irq=%d #autofd=%d \n",
         sbuf, cbuf, eyecam->stat_irq, eyecam->stat_autofd);
}

static void
eyecam_irq(int irq, void *dev, struct pt_regs *regs)
{
  struct eyecam_device *eyecam = dev;

  eyecam->stat_irq++;
  eyecam->stat_control = parport_read_control(eyecam->pport);
  eyecam->stat_status  = parport_read_status(eyecam->pport);
  eyecam_print_stat(eyecam);
#if driver_working
  if (NULL != eyecam->framebuf)
    eyecam->framebuf[eyecam->framecount++] = parport_read_data(eyecam->pport);
  if (eyecam->framecount > eyecam->framesize)
    {
      eyecam->capture_ready = 1;
      wake_up_interruptible(&eyecam->capture_wait);
    }
#endif
}

static int
eyecam_open(struct video_device *dev, int flags)
{
  struct eyecam_device *eyecam = (struct eyecam_device *)dev;

  printk(KERN_ERR "eyecam: open() users=%d\n", eyecam->users);

  if (eyecam->users)
    return -EBUSY;

  parport_claim_or_block(eyecam->pdev);

  /* Enable interrupt generation on ACK moving from logical 1 to 0. */
#if 1
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_INTEN,
                         PARPORT_CONTROL_INTEN);
#else
  eyecam->pport->ops->enable_irq(eyecam->pport);
#endif

  eyecam->users++;
  MOD_INC_USE_COUNT;
  return 0;
}

static void
eyecam_close(struct video_device *dev)
{
  struct eyecam_device *eyecam = (struct eyecam_device *)dev;

  printk(KERN_ERR "eyecam: close() users=%d\n", eyecam->users);

  eyecam->users--;
  parport_release(eyecam->pdev);
  MOD_DEC_USE_COUNT;
}

static unsigned int
eyecam_poll(struct video_device *dev, struct file *file,
            poll_table *wait)
{
  struct eyecam_device *eyecam = (struct eyecam_device *)dev;

  printk(KERN_ERR "eyecam: poll() users=%d\n", eyecam->users);

  poll_wait(file, &eyecam->capture_wait, wait);
  if (eyecam->capture_ready)
    return POLLIN|POLLRDNORM;
  return 0;
}

static long
eyecam_read(struct video_device *dev, char *buf, unsigned long count,
            int noblock)
{
  struct eyecam_device *eyecam = (struct eyecam_device *)dev;
  struct wait_queue wait = { current, NULL};
  u8 *ptr;
  unsigned int len;
  unsigned int i;

  printk(KERN_ERR "eyecam: read() buf=0x%p count=%ld noblock=%s\n",
         buf, count, noblock ? "yes" : "no");

#if 1 || driver_not_working
  add_wait_queue(&eyecam->capture_wait, &wait);

  while (!eyecam->capture_ready)
    {
      eyecam_print_stat(eyecam);

      if (noblock)
        {
          remove_wait_queue(&eyecam->capture_wait, &wait);
          current->state = TASK_RUNNING;
          return -EWOULDBLOCK;
        }
      if (signal_pending(current))
        {
          remove_wait_queue(&eyecam->capture_wait, &wait);
          current->state = TASK_RUNNING;
          return -ERESTARTSYS;
        }
      schedule();
      current->state = TASK_INTERRUPTIBLE;
    }
  remove_wait_queue(&eyecam->capture_wait, &wait);
  current->state = TASK_RUNNING;
#else
  {
    int i, j;
    for (i = 0; i < 5; i++) {
      eyecam_print_stat(eyecam);
      eyecam->stat_control ^= PARPORT_CONTROL_AUTOFD;
      parport_write_control(eyecam->pport, eyecam->stat_control);
      eyecam->stat_autofd++;

#if 0
      schedule_timeout(HZ*1000); /* Sleep for +- 10 mS XXX check this*/
#else
      for (j = 0; j < 1000; j++)
        schedule(); /* Sleep for +- 10 mS */
#endif

      eyecam->stat_status  = parport_read_status(eyecam->pport);
    }
  }
#endif
  eyecam->capture_ready = 0;

  ptr = (u8*)buf;

  /* 24 bit RGB */
  len = eyecam->capture_fb.width * 3 * eyecam->capture_fb.height;

  if (len > count)
    len = count; /* Doesn't all fit */

  for (i=0; i<len; i++)
    {
      put_user(eyecam->framebuf[i], ptr);
      ptr++;
    }
  hardware_restart_capture(eyecam);

  return i;
}

static int
eyecam_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
  struct eyecam_device *eyecam = (struct eyecam_device *)dev;
  switch (cmd)
    {
    case VIDIOCGCAP:
      {
        struct video_capability v;
        v.type = VID_TYPE_CAPTURE;
        v.channels = 1;
        v.audios = 0;
        v.maxwidth = 82;
        v.minwidth = 80;
        v.maxheight = 62;
        v.minheight = 60;
        strcpy(v.name, EYECAM);
        if (copy_to_user(arg, &v, sizeof(v)))
          return -EFAULT;
        return 0;
        break;
      }
    case VIDIOCGCHAN:
      {
        struct video_channel v;
        if (copy_from_user(&v, arg, sizeof(v)))
          return -EFAULT;
        if (v.channel != 0)
          return -EINVAL;
        v.flags = 0;
        v.tuners = 0;
        v.type = VIDEO_TYPE_CAMERA;
        v.norm = VIDEO_MODE_AUTO;
        strcpy(v.name, "Camera Input");
        if (copy_to_user(&v, arg, sizeof(v)))
          return -EFAULT;
        return 0;
        break;
      }
    case VIDIOCSCHAN:
      {
        struct video_channel v;
        if (copy_from_user(&v, arg, sizeof(v)))
          return -EFAULT;
        if (v.channel != 0)
          return -EINVAL;
        if (v.norm != VIDEO_MODE_AUTO)
          return -EINVAL;
        return 0;
        break;
      }
    case VIDIOCGPICT:
      {
        struct video_picture v;
        v.brightness = hardware_brightness(eyecam);
        v.hue = hardware_hue(eyecam);
        v.colour = hardware_saturation(eyecam);
        v.contrast = hardware_contrast(eyecam);
        v.whiteness = 32768; /* Not settable */
        v.depth = 24; /* 24bit */
        v.palette = VIDEO_PALETTE_RGB24;
        if (copy_to_user(&v, arg, sizeof(v)))
          return -EFAULT;
        return 0;
        break;
      }
    case VIDIOCSPICT:
      {
        struct video_picture v;
        if (copy_from_user(&v, arg, sizeof(v)))
          return -EFAULT;
        if (v.depth != 24 || v.palette != VIDEO_PALETTE_RGB24)
          return -EINVAL;
        hardware_set_brightness(eyecam, v.brightness);
        hardware_set_hue(eyecam, v.hue);
        hardware_set_saturation(eyecam, v.colour);
        hardware_set_contrast(eyecam, v.contrast);
        return 0;
        break;
      }
    case VIDIOCGFBUF:
      {
        if (copy_to_user(arg, &eyecam->capture_fb, sizeof(eyecam->capture_fb)))
          return -EFAULT;
        return 0;
        break;
      }
    case VIDIOCSFBUF:
      {
        struct video_buffer v;
        if (copy_from_user(&v, arg, sizeof(v)))
          return -EFAULT;

        /* Only support 80x60 and 82x62 */
        if ( ! ( 80 == v.width && 60 == v.height) ||
             ! ( 82 == v.width && 62 == v.height) )
          return -EINVAL;

        memcpy(&eyecam->capture_fb, &v, sizeof(v));
        hardware_set_fb(eyecam, &v);
        return 0;
        break;
      }
    case VIDIOCCAPTURE:
      {
        int v;
        if (get_user(v, (int*)arg))
          return -EFAULT;
        if (0 == v)
          hardware_capture_off(eyecam);
        else
          {
            if (0 == eyecam->capture_fb.width)
              return -EINVAL;
            hardware_capture_on(eyecam);
          }
        return 0;
        break;
      }
    default:
      return -ENOIOCTLCMD;
      break;
    }
}

static struct video_device eyecam_template =
{
  EYECAM,
  VID_TYPE_CAPTURE,
  VID_HARDWARE_EYECAM,
  eyecam_open,
  eyecam_close,
  eyecam_read,
  NULL, /* no write */
#if LINUX_VERSION_CODE >= 0x020100
  eyecam_poll,
#endif
  eyecam_ioctl,
  NULL, /* no mmap */
  NULL, /* no special init function */
  NULL, /* no private data */
  0, 0  /* Not sure what busy and minor is */
};

#if newer_kernel_2_4
static void
eyecam_attach(struct parport *port)
{
  printk(KERN_ERR "attach: Checking %s\n", port->name);
  eyecam_init_camera(port);
}

static void
eyecam_detach(struct parport *)
{
  printk(KERN_ERR "detach: Loosing %s\n", port->name);
}

struct parport_driver eyecam_driver =
{
  EYECAM,
  eyecam_attach,
  eyecam_detach,
  NULL
};
#endif

#define BANNER "Eyecam (VV6300) digital camera driver v0.01\n"

static int
eyecam_init_camera(struct parport *port)
{
  struct eyecam_device *eyecam;

  if (eyecam_num_cams == MAX_CAMS)
    {
      printk(KERN_ERR "Too many Eyecams (max %d)\n", MAX_CAMS);
      return -ENOSPC;
    }

  eyecam = kmalloc(sizeof(*eyecam), GFP_KERNEL);
  if (NULL == eyecam)
    return -ENODEV;

  eyecam->pport = port;
  eyecam->pdev = parport_register_device(port, "eyecam", NULL, NULL,
                                         eyecam_irq, 0, eyecam);

  if (NULL == eyecam->pdev)
    {
      printk(KERN_ERR "eyecam: couldn't register for %s.\n", port->name);
      kfree(eyecam);
      return -ENODEV;
    }

  memcpy(&eyecam->vdev, &eyecam_template, sizeof(eyecam_template));

  eyecam->users = 0;
  eyecam->capture_wait = NULL;
  eyecam->capture_ready = 0;
  eyecam->framesize = 0;
  eyecam->framecount = 0;
  eyecam->framebuf = NULL;
  memset(&eyecam->capture_fb, 0, sizeof(eyecam->capture_fb));

  eyecam_init_stat(eyecam);

  parport_claim_or_block(eyecam->pdev);

  if (0 == hardware_detect(eyecam))
    {
      printk(KERN_ERR "eyecam: No camera found on %s\n", port->name);
      parport_release(eyecam->pdev);
      parport_unregister_device(eyecam->pdev);
      kfree(eyecam);
      return -ENODEV;
    }

  hardware_setup(eyecam);

  parport_release(eyecam->pdev);

  printk(KERN_ERR "Eyecam found on %s\n", eyecam->pport->name);

  if (-1 == video_register_device(&eyecam->vdev, VFL_TYPE_GRABBER))
    {
      parport_unregister_device(eyecam->pdev);
      kfree(eyecam);
      return -EINVAL;
    }
  eyecams[eyecam_num_cams++] = eyecam;
  return 0;
}

static int __init
eyecam_init(struct video_init *v)
{
  struct parport *port;
  int i;

  for (i = 0; i < MAX_CAMS; i++)
    eyecams[i] = NULL;

  printk(BANNER);
#if newer_kernel_2_4
  parport_register_driver(eyecam_driver);
#else
  for (port = parport_enumerate(); port; port=port->next)
    if (0 != (i = eyecam_init_camera(port)))
      return i;
#endif

  if (eyecam_num_cams <= 0)
    return -ENODEV;

  return 0;
}

static void
eyecam_cleanup(struct eyecam_device *eyecam)
{
  video_unregister_device(&eyecam->vdev);
  parport_unregister_device(eyecam->pdev);
  kfree(eyecam);
}

#ifdef MODULE

#if LINUX_VERSION_CODE >= 0x020117
MODULE_AUTHOR("Petter Reinholdtsen <pere@td.org.uit.no>");
MODULE_DESCRIPTION("A parport driver for the EyeCam (VV6300) digital camera.");
MODULE_PARM(maxpoll,"i");
MODULE_PARM(yieldlines,"i");
#endif

EXPORT_NO_SYMBOLS;

int __init
init_module(void)
{
  printk(KERN_ERR "eyecam: init_module()\n");
  return eyecam_init(NULL);
}

void
cleanup_module(void)
{
  int i;
  printk(KERN_ERR "eyecam: cleanup_module()\n");
  for (i = 0; i < MAX_CAMS; i++)
    {
      if (NULL != eyecams[i])
        eyecam_cleanup(eyecams[i]);
      eyecams[i] = NULL;
    }
}

#endif

/**************************** HW support *********************************/

/* time (microseconds) to wait between driving clock transitions for SCL */
#define CLK_DELAY 100

/* codes to send to camera for read and write operations */
#define CAM_READ 0x21
#define CAM_WRITE 0x20

/* camera register addresses */
#define SETUP_0  0x10
#define SETUP_1  0x11
#define SETUP_2  0x12
#define SETUP_3  0x13
#define SETUP_4  0x14
#define F_EXP    0x21
#define C_EXP    0x23
#define GAIN     0x24
#define CLK_DIV  0x25
#define GAIN_LIM 0x26
#define EXP_LT   0x27
#define EXP_CT   0x28
#define EXP_HT   0x29
#define FRM_AVG  0x73
#define DC_THRES 0x75

/* If this is false, we never check if the camera respond to your commands */
static int uart_twoway = 0;

/* Use this to avoid buffered value */
__inline__ unsigned char
parport_read_control_direct(struct parport *p)
{
  return inb(CONTROL(p));
}

/* SLCTIN pin 17 */
static inline void
clear_sda(struct eyecam_device *eyecam)
{
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_SELECT, 0);
}

static inline void
set_sda(struct eyecam_device *eyecam)
{
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_SELECT,
                         PARPORT_CONTROL_SELECT);
}

static inline char
read_sda(struct eyecam_device *eyecam)
{
  eyecam->stat_control =  parport_read_control_direct(eyecam->pport);
  return eyecam->stat_control & PARPORT_CONTROL_SELECT ? 0 : 1;
}

/* STROBE pin 1 */
static inline void
clear_scl(struct eyecam_device *eyecam)
{
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_STROBE, 0);
}

static inline void
set_scl(struct eyecam_device *eyecam)
{
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_STROBE,
                         PARPORT_CONTROL_STROBE);
}

static void
set_input_mode(struct eyecam_device *eyecam)
{
  /* Go to mode 001; PS2 */
  parport_pc_frob_econtrol (eyecam->pport, 0xe0, 0x20);

  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_DIRECTION,
                         PARPORT_CONTROL_DIRECTION);
}

static void
set_output_mode(struct eyecam_device *eyecam)
{
  eyecam->stat_control =
    parport_frob_control(eyecam->pport, PARPORT_CONTROL_DIRECTION, 0);
}

/* BUSY pin 11 */
static inline u8
read_fst(struct eyecam_device *eyecam)
{
  eyecam->stat_status =  parport_read_status(eyecam->pport);
  return eyecam->stat_status & PARPORT_STATUS_BUSY ? 1 : 0;
}

/* ACK pin 10 */
static inline u8
read_qck(struct eyecam_device *eyecam)
{
  eyecam->stat_status =  parport_read_status(eyecam->pport);
  return eyecam->stat_status & PARPORT_STATUS_ACK ? 1 : 0;
}

static void
start_condition(struct eyecam_device *eyecam)
{
    set_sda(eyecam);
    set_scl(eyecam);
    udelay(CLK_DELAY);
    clear_sda(eyecam);
    udelay(CLK_DELAY);
    clear_scl(eyecam);
}

static void
stop_condition(struct eyecam_device *eyecam)
{
    set_scl(eyecam);
    set_sda(eyecam);
    udelay(CLK_DELAY);
}

static int
send_byte(struct eyecam_device *eyecam, u8 val)
{
  int retval = 0;
  int i;
  char sbuf[9], cbuf[9];

  int2str(sbuf, parport_read_status(eyecam->pport), 2, 8, 1);
  int2str(cbuf, parport_read_control(eyecam->pport), 2, 8, 1);

  printk(KERN_ERR "eyecam: send_byte(0x%02x) s=%s c=%s\n", val, sbuf, cbuf);

  for (i=0; i<8; i++) {
    if ((val & 0x80) == 0) {
      clear_sda(eyecam);
    } else {
      set_sda(eyecam);
    }

    udelay(CLK_DELAY);
    set_scl(eyecam);
    udelay(CLK_DELAY);
    clear_scl(eyecam);
    udelay(CLK_DELAY);

    val <<= 1;

    int2str(sbuf, parport_read_status(eyecam->pport), 2, 8, 1);
    int2str(cbuf, parport_read_control(eyecam->pport), 2, 8, 1);
    printk(KERN_ERR "eyecam: send_byte() %2d  s=%s c=%s\n", val&0x80 ? 1 : 0,
           sbuf, cbuf);
  }

  set_scl(eyecam);
  udelay(CLK_DELAY);
  clear_sda(eyecam); /* Make sure pin is +5V to prepare for input */
  if (uart_twoway && read_sda(eyecam))
    retval = -1;
  clear_scl(eyecam);

  int2str(sbuf, parport_read_status(eyecam->pport), 2, 8, 1);
  int2str(cbuf, parport_read_control(eyecam->pport), 2, 8, 1);
  printk(KERN_ERR "eyecam: send_byte()=%2d  s=%s c=%s\n", retval, sbuf, cbuf);

  return retval;
}

static int
write_message(struct eyecam_device *eyecam, u8 index, u8 data)
{
  start_condition(eyecam);
  if (0 != send_byte(eyecam, CAM_WRITE))
    {
      printk(KERN_ERR "eyecam: write_message fail1\n");
      return -1;
    }
  if (0 != send_byte(eyecam, index))
    {
      printk(KERN_ERR "eyecam: write_message fail2\n");
      return -1;
    }
  if (0 != send_byte(eyecam, data))
    {
      printk(KERN_ERR "eyecam: write_message fail3\n");
      return -1;
    }
  stop_condition(eyecam);
  return 0;
}

static int
hardware_detect(struct eyecam_device *eyecam)
{
  printk(KERN_ERR "eyecam: detecting on %s\n", eyecam->pport->name);

  /* return 1; */

  set_input_mode(eyecam);

  /* dummy message for startup */
  write_message(eyecam, 0xff, 0x00);

  if (0 != write_message(eyecam, SETUP_0, 0x07))
    {
      printk(KERN_ERR "eyecam: SETUP_0 failed.\n");
      return 0;
    }
  if (0 != write_message(eyecam, SETUP_1, 0x60))
    {
      printk(KERN_ERR "eyecam: SETUP_1 failed.\n");
      return 0;
    }
  if (0 != write_message(eyecam, SETUP_4, 0x4c))
    {
      printk(KERN_ERR "eyecam: SETUP_4 failed.\n");
      return 0;
    }
  if (0 != write_message(eyecam, CLK_DIV, 0x03))
    {
      printk(KERN_ERR "eyecam: CLK_DIV failed.\n");
      return 0;
    }
  set_input_mode(eyecam);
  return 1;
};


/**********************************************************************/
/** Convert from int to string using base.
 */
int
int2str(char *dst, int value, int base, int len, int zerofill)
{
  const char values[] = "0123456789ABCDEF";
  char buf[10];
  int rest, digits = 0;
  int negative = 0;

  if (base < 2 || base > 16)
    return -1;

  if (0 > value)
    {
      negative = 1;
      value *= -1;
    }

  do
    {
      int s = value % base;
      buf[digits++] = values[s];
      value = (value-s)/base;
    }
  while (value);

  if (negative)
    buf[digits++] = '-';

  buf[digits] = 0;

  if (0 == len)
    len = digits;

  if (len < digits)
    return -1; /* Not enough space */

  rest = len;
  while (rest)
    {
      if (rest > digits)
        {
          if (zerofill && !negative)
            dst[len-rest] = '0';
          else
            dst[len-rest] = ' ';
        }
      else
        dst[len-rest] = buf[rest-1];
      rest--;
    }
  dst[len] = 0;
  return len;
}
